import boto3
import json
import logging
import pytz
import os
import hashlib

from abc import ABC
from datetime import datetime
from backup_reporter.dataclass import BackupMetadata, BackupFileInfo
from backup_reporter.utils import exec_cmd
from fnmatch import fnmatch


class BackupReporter(ABC):
    '''
        Base backup reporter with common functionality.
        It is highly recommended not to rewrite methods (except _gather_metadata) in child classes.
    '''
    def __init__(
            self,
            aws_access_key_id: str,
            aws_secret_access_key: str,
            aws_region: str,
            s3_path: str,
            type: str,
            customer: str,
            supposed_backups_count: str,
            description: str,
            aws_endpoint_url: str = None) -> None:
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.aws_region = aws_region
        self.aws_endpoint_url = aws_endpoint_url
        self.s3_path = s3_path

        self.metadata = BackupMetadata()
        self.metadata.type = type
        self.metadata.customer = customer
        self.metadata.supposed_backups_count = supposed_backups_count
        self.metadata.description = description

    def _gather_metadata(self) -> BackupMetadata:
        '''
            Gather information about backup to dict of variables
            This method exists to be overridden in childs classes
        '''
        raise Exception('Method _gather_metadata must be overwritten in child class')

    def _upload_metadata(self, metadata: BackupMetadata) -> None:
        '''Upload metadata file to place, where backups stored'''
        logging.info(f"Uploud metadata to {self.s3_path} ...")
        kwargs = {
           "aws_access_key_id": self.aws_access_key_id,
           "aws_secret_access_key": self.aws_secret_access_key,
           "region_name": self.aws_region,
           "endpoint_url": self.aws_endpoint_url
        }
        s3 = boto3.resource(
            's3',
            **{k:v for k,v in kwargs.items() if v is not None}
        )
        metadata_file_name = "/".join(self.s3_path.split("/")[3:])
        s3_path = self.s3_path.split("/")[2]
        s3.Object(s3_path, metadata_file_name).put(Body=str(metadata))
        logging.info(f"Uploud metadata success")

    def report(self) -> None:
        '''Check backup status, compile it to json metadata file and upload'''
        metadata = self._gather_metadata()
        self._upload_metadata(metadata)


class DockerPostgresBackupReporter(BackupReporter):
    '''
        Reporter for Postgresql running in containers.
        For working reporter require permissions to work with docker socket.
    '''
    def __init__(
            self,
            container_name: str,
            aws_access_key_id: str,
            aws_secret_access_key: str,
            aws_region: str,
            s3_path: str,
            customer: str,
            supposed_backups_count: str,
            description: str,
            aws_endpoint_url: str = None) -> None:

        super().__init__(
            aws_access_key_id = aws_access_key_id,
            aws_secret_access_key = aws_secret_access_key,
            aws_region = aws_region,
            s3_path = s3_path,
            customer = customer,
            supposed_backups_count = supposed_backups_count,
            type = "DockerPostgres",
            description = description,
            aws_endpoint_url = aws_endpoint_url)

        self.container_name = container_name
        self.metadata.last_backup_date = None

    def _gather_metadata(self) -> BackupMetadata:
        '''Gather information about backup to dict of variables'''
        logging.info(f"Gather metadata from {self.container_name} ...")
        wal_show = exec_cmd(["docker", "exec", "-i", self.container_name, "wal-g", "wal-show", "--detailed-json" ])
        wal_show = json.loads(wal_show)
        full_backup_count = 0
        last_full_backup_date = None
        incremental_backup_count = 0
        for backup in wal_show[0]['backups']:
            backup_time = datetime.strptime(backup.get('time'), '%Y-%m-%dT%H:%M:%SZ')
            backup_time = backup_time.replace(tzinfo=pytz.UTC, microsecond=0)
            if not self.metadata.last_backup_date or backup_time > self.metadata.last_backup_date:
                self.metadata.last_backup_date = backup_time  # Beware, this is ALWAYS about LAST backup - full or incremental
                self.metadata.backup_name = backup.get("backup_name", "None")  # Also ALWAYS about LAST backup
                self.metadata.size = round(backup.get("compressed_size", "None") /1024/1024, 1)  # Can be overridden below
                finish_time = datetime.strptime(backup.get('finish_time'), backup.get('date_fmt'))  # Can be overridden below
                start_time = datetime.strptime(backup.get('start_time'), backup.get('date_fmt'))  # Can be overridden below
                self.metadata.time = str(finish_time - start_time)  # Can be overridden below

            backup_wal_file_name = backup.get("wal_file_name", "Unknown")
            if backup['backup_name'].endswith(backup_wal_file_name):  # If so, we're looking at full backup
                full_backup_count += 1
                if not last_full_backup_date or backup_time > last_full_backup_date:  # Override backup info with the size of last full backup
                    last_full_backup_date = backup_time
                    self.metadata.size = round(backup.get("compressed_size", "None") /1024/1024, 1)
                    finish_time = datetime.strptime(backup.get('finish_time'), backup.get('date_fmt'))  # Can be overridden below
                    start_time = datetime.strptime(backup.get('start_time'), backup.get('date_fmt'))  # Can be overridden below
                    self.metadata.time = str(finish_time - start_time)  # Can be overridden below
            else:
                incremental_backup_count += 1
        # Now we have to serialize last backup date
        self.metadata.last_backup_date = str(self.metadata.last_backup_date)

        self.metadata.count_of_backups = f"{len(wal_show[0]['backups'])} total / {full_backup_count} full / {incremental_backup_count} incremental"

        s3_path = "/".join(self.s3_path.split("/")[:3])
        self.metadata.placement = s3_path

        logging.info("Gather metadata success")
        logging.debug(self.metadata)
        return self.metadata


class FilesBucketReporterBackupReporter(BackupReporter):
    '''
        Report about backups from S3 bucket with plain files. Usually they are 1 file per 1 backup, but different schemes are available.
    '''
    def __init__(
            self,
            aws_access_key_id: str,
            aws_secret_access_key: str,
            aws_region: str,
            s3_path: str,
            customer: str,
            supposed_backups_count: str,
            description: str,
            files_mask: str,
            aws_endpoint_url: str = None) -> None:

        super().__init__(
            aws_access_key_id = aws_access_key_id,
            aws_secret_access_key = aws_secret_access_key,
            aws_region = aws_region,
            s3_path = s3_path,
            customer = customer,
            supposed_backups_count = supposed_backups_count,
            type = "FilesBucket",
            description = description,
            aws_endpoint_url = aws_endpoint_url)

        self.metadata.last_backup_date = None
        self.files_mask = files_mask

    def _gather_metadata(self) -> BackupMetadata:
        '''
            Gather information about backup from files in S3
        '''
        kwargs = {
           "aws_access_key_id": self.aws_access_key_id,
           "aws_secret_access_key": self.aws_secret_access_key,
           "region_name": self.aws_region,
           "endpoint_url": self.aws_endpoint_url
        }
        s3 = boto3.resource(
            's3',
            **{k:v for k,v in kwargs.items() if v is not None}
        )

        bucket_name = self.s3_path.split("/")[2]
        s3 = s3.Bucket(bucket_name)

        latest_backup = {"key": None, "last_modified": datetime(2000, 1, 1, tzinfo=pytz.UTC), "size": 0} # Default latest backup
        count_of_backups = 0
        # Get latest backup file
        for object in s3.objects.all():
            if fnmatch(object.key, self.files_mask): # Check if object name matches with files mask from config file
                if latest_backup["last_modified"] < object.last_modified:
                    latest_backup = {"key": object.key, "last_modified": object.last_modified, "size": object.size}
                self.metadata.backups.append(BackupFileInfo(
                    size=round(object.size/1024/1024, 1),
                    backup_date=object.last_modified,
                    backup_name=object.key
                ))
                count_of_backups += 1

        self.metadata.count_of_backups = count_of_backups
        self.metadata.last_backup_date = latest_backup["last_modified"]
        self.metadata.backup_name = latest_backup["key"]
        self.metadata.placement = bucket_name
        self.metadata.size = round(latest_backup["size"]/1024/1024, 1)
        self.metadata.time = 0

        return self.metadata

class FilesReporterBackupReporter(BackupReporter):
    '''
        Report about backups from plain files. Usually they are 1 file per 1 backup, but different schemes are available.
    '''
    def __init__(
            self,
            aws_access_key_id: str,
            aws_secret_access_key: str,
            aws_region: str,
            s3_path: str,
            customer: str,
            supposed_backups_count: str,
            description: str,
            files_mask: str,
            backups_dir: str,
            aws_endpoint_url: str = None) -> None:

        super().__init__(
            aws_access_key_id = aws_access_key_id,
            aws_secret_access_key = aws_secret_access_key,
            aws_region = aws_region,
            s3_path = s3_path,
            customer = customer,
            supposed_backups_count = supposed_backups_count,
            type = "Files",
            description = description,
            aws_endpoint_url = aws_endpoint_url)

        self.files_mask = files_mask
        self.backups_dir = backups_dir

    def _gather_metadata(self) -> BackupMetadata:
        '''
            Gather information about backup from files in the host
        '''
        latest_backup = {"key": None, "last_modified": datetime(2000, 1, 1, tzinfo=pytz.UTC), "size": 0} # Default latest backup
        count_of_backups = 0
        # Get latest backup file
        for dirpath, _, filenames in os.walk(self.backups_dir):
            for filename in filenames:
                if fnmatch(filename, self.files_mask):
                    count_of_backups += 1
                    filepath = os.path.join(dirpath, filename)
                    try:
                        stat = os.stat(filepath)
                    except FileNotFoundError:
                        continue

                    sha1 = hashlib.sha1()
                    try:
                        with open(filepath, "rb") as f:
                            while chunk := f.read(8192):
                                sha1.update(chunk)
                        sha1sum = sha1.hexdigest()
                    except Exception as e:
                        logging.error(f"Can not read the file!")

                    mtime = datetime.fromtimestamp(stat.st_mtime, tz=pytz.UTC).replace(microsecond=0)
                    if mtime > latest_backup["last_modified"]:
                        latest_backup = {"key": filename, "last_modified": mtime, "size": stat.st_size, "sha1sum": sha1sum}

                    self.metadata.backups.append(BackupFileInfo(
                        size=round(stat.st_size/1024/1024, 1),
                        backup_date=mtime,
                        backup_name=filename,
                        sha1sum=sha1sum
                    ))

        if latest_backup["key"]: # If at least one file exists
            self.metadata.count_of_backups = count_of_backups
            self.metadata.last_backup_date = latest_backup["last_modified"]
            self.metadata.backup_name = latest_backup["key"]
            self.metadata.placement = self.backups_dir
            self.metadata.size = round(latest_backup["size"]/1024/1024, 1)
            self.metadata.time = 0
            self.metadata.sha1sum = latest_backup["sha1sum"]

        logging.info("Gather metadata success")
        logging.debug(self.metadata)
        return self.metadata

class S3MariadbBackupReporter(BackupReporter):
    '''
        Reporter for uploaded to S3 files with backups.
    '''
    def __init__(
            self,
            aws_access_key_id: str,
            aws_secret_access_key: str,
            aws_region: str,
            s3_path: str,
            customer: str,
            supposed_backups_count: str,
            description: str,
            aws_endpoint_url: str = None) -> None:

        super().__init__(
            aws_access_key_id = aws_access_key_id,
            aws_secret_access_key = aws_secret_access_key,
            aws_region = aws_region,
            s3_path = s3_path,
            customer = customer,
            supposed_backups_count = supposed_backups_count,
            type = "DockerMariadb",
            description = description,
            aws_endpoint_url = aws_endpoint_url)

        self.metadata.last_backup_date = None

    def _gather_metadata(self) -> BackupMetadata:
        kwargs = {
            "aws_access_key_id": self.aws_access_key_id,
            "aws_secret_access_key": self.aws_secret_access_key,
            "region_name": self.aws_region,
            "endpoint_url": self.aws_endpoint_url
        }
        s3 = boto3.client(
            's3',
            **{k:v for k,v in kwargs.items() if v is not None}
        )

        bucket_name = self.s3_path.split("/")[2]
        directories = []
        count_of_backups = 0
        backup_total_size = 0
        response = s3.list_objects_v2(Bucket=bucket_name, Prefix='mariadb/full/', Delimiter='/')
        if 'CommonPrefixes' in response:
            directories = [prefix['Prefix'] for prefix in response['CommonPrefixes']]
            latest_full_backup = directories[-1]
            latest_date_of_backup = latest_full_backup.split('/')[-2]
            count_of_backups = len(directories)
            inc_path = 'mariadb/inc/' + latest_date_of_backup
            response_from_inc_path = s3.list_objects_v2(Bucket=bucket_name, Prefix=inc_path, Delimiter='/')
            if 'CommonPrefixes' in response_from_inc_path:
                directories = [prefix['Prefix'] for prefix in response_from_inc_path['CommonPrefixes']]
                latest_backup = directories[-1]
                latest_date_of_backup = latest_backup.split('/')[-2]
            else:
                logging.info("No directories found in the incremental path.")
                latest_backup = latest_full_backup
            objects_of_backup = s3.list_objects_v2(Bucket=bucket_name, Prefix=latest_backup)
            if 'Contents' in objects_of_backup:
                for obj in objects_of_backup['Contents']:
                    backup_total_size += obj['Size']
        else:
            logging.info("No directories found in the specified path.")
            latest_backup = 'None'
            count_of_backups = 0
            backup_total_size = 0
        self.metadata.count_of_backups = count_of_backups
        self.metadata.last_backup_date = latest_date_of_backup
        self.metadata.backup_name = latest_backup
        self.metadata.placement = bucket_name
        self.metadata.size = round(backup_total_size/1024/1024, 1)
        self.metadata.time = 0
        return self.metadata



import logging
import subprocess

from argparse import Namespace
from yaml import safe_load
from mergedeep import merge


def exec_cmd(args: list) -> str:
    # Exec input command
    out = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    stdout, stderr = out.communicate()

    if out.returncode != 0:
        stdout_msg = stdout.decode('utf-8') if stdout is not None else ''
        stderr_msg = stderr.decode('utf-8') if stderr is not None else ''
        logging.info(out.returncode)
        raise Exception(f"Command returned code {out.returncode}. Stdout: '{stdout_msg}' Stderr: '{stderr_msg}'")

    return stdout.decode("utf-8")


def set_confs(args: Namespace) -> dict:
    confs = {}
    if args.config != "":
        with open(args.config) as conf:
            confs = safe_load(conf)

    confs = merge(args.__dict__, confs)

    return confs



import json

from dataclasses import dataclass, asdict, field

@dataclass
class BackupFileInfo:
    size: str = None
    backup_date: str = None
    backup_name: str = None
    sha1sum: str = None
    def __str__(self):
        '''String representation of that DataClass is valid json string'''
        return json.dumps(asdict(self), default=str)

@dataclass
class BackupMetadata:
    '''Class contain fields with info about backup'''
    type: str = None
    size: str = None
    time: str = None
    customer: str = None
    placement: str = None
    backup_name: str = None
    description: str = None
    last_backup_date: str = None
    count_of_backups: str = None
    supposed_backups_count: str = None
    sha1sum: str = None
    backups: list[BackupFileInfo] = field(default_factory=list)
    def __str__(self):
        '''String representation of that DataClass is valid json string'''
        return json.dumps(asdict(self), default=str)



import argparse
import sys
import logging
import backup_reporter.reporters as rps

from backup_reporter.collector import BackupCollector
from backup_reporter.utils import set_confs


def start():
    arg_parser = argparse.ArgumentParser()

    arg_parser.add_argument("--config",
        type=str,
        default="",
        help="Set config path"
    )

    arguments = arg_parser.parse_known_args()[0]
    confs = set_confs(arguments)

    logging.basicConfig(
        level=getattr(logging, confs.get("logging_level", "INFO")),
        handlers=[logging.StreamHandler(sys.stdout)]
    )

    if confs["collector"]:
        logging.info("Run collector")
        collector = BackupCollector(buckets = confs.get('bucket', None),
            google_spreadsheet_credentials_path = confs.get('google_spreadsheet_credentials_path', None),
            spreadsheet_name = confs.get('spreadsheet_name', None),
            worksheet_name = confs.get('worksheet_name', None),
            sheet_owner = confs.get('sheet_owner', None))
        collector.collect()

    elif confs["docker_postgres"]:
        logging.info("Report about docker-postgres backups")
        reporter = rps.DockerPostgresBackupReporter(
            aws_access_key_id = confs["bucket"][0].get("aws_access_key_id", None),
            aws_secret_access_key = confs["bucket"][0].get("aws_secret_access_key", None),
            aws_region = confs["bucket"][0].get("aws_region", None),
            s3_path = confs["bucket"][0].get("s3_path", None),
            container_name = confs.get("container_name", None),
            customer = confs.get("customer", None),
            supposed_backups_count = confs.get("supposed_backups_count", None),
            aws_endpoint_url = confs["bucket"][0].get("aws_endpoint_url", None),
            description = confs.get("description", None)
        )
        reporter.report()

    elif confs["files_bucket"]:
        logging.info("Report about files backups in S3 buckets")
        reporter = rps.FilesBucketReporterBackupReporter(
            aws_access_key_id = confs["bucket"][0].get("aws_access_key_id", None),
            aws_secret_access_key = confs["bucket"][0].get("aws_secret_access_key", None),
            aws_region = confs["bucket"][0].get("aws_region", None),
            s3_path = confs["bucket"][0].get("s3_path", None),
            customer = confs.get("customer", None),
            supposed_backups_count = confs.get("supposed_backups_count", None),
            aws_endpoint_url = confs["bucket"][0].get("aws_endpoint_url", None),
            description = confs.get("description", None),
            files_mask = confs.get("files_mask", None)
        )
        reporter.report()

    elif confs["files"]:
        logging.info("Report about files backups in the host")
        reporter = rps.FilesReporterBackupReporter(
            aws_access_key_id = confs["bucket"][0].get("aws_access_key_id", None),
            aws_secret_access_key = confs["bucket"][0].get("aws_secret_access_key", None),
            aws_region = confs["bucket"][0].get("aws_region", None),
            s3_path = confs["bucket"][0].get("s3_path", None),
            customer = confs.get("customer", None),
            supposed_backups_count = confs.get("supposed_backups_count", None),
            aws_endpoint_url = confs["bucket"][0].get("aws_endpoint_url", None),
            description = confs.get("description", None),
            files_mask = confs.get("files_mask", None),
            backups_dir = confs.get("backups_dir", None)
        )
        reporter.report()

    elif confs["s3_mariadb"]:
        logging.info("Report about S3-mariadb backups")
        reporter = rps.S3MariadbBackupReporter(
            aws_access_key_id = confs["bucket"][0].get("aws_access_key_id", None),
            aws_secret_access_key = confs["bucket"][0].get("aws_secret_access_key", None),
            aws_region = confs["bucket"][0].get("aws_region", None),
            s3_path = confs["bucket"][0].get("s3_path", None),
            customer = confs.get("customer", None),
            supposed_backups_count = confs.get("supposed_backups_count", None),
            aws_endpoint_url = confs["bucket"][0].get("aws_endpoint_url", None),
            description = confs.get("description", None),
        )
        reporter.report()

    else:
        logging.info("You MUST choose either reporter mode or collector mode")
        exit(1)

if __name__ == "__main__":
    start()



import os
import csv
import json
import boto3
import gspread
import logging
import datetime
import dateparser
from time import sleep
from gspread_formatting import Color, CellFormat, format_cell_range
from oauth2client.service_account import ServiceAccountCredentials

from backup_reporter.dataclass import BackupMetadata, BackupFileInfo


class BackupCollector:
    def __init__(self, buckets: list,
            google_spreadsheet_credentials_path: str,
            spreadsheet_name: str,
            worksheet_name: str,
            sheet_owner: str) -> None:
        self.buckets = buckets
        self.credentials_path = google_spreadsheet_credentials_path
        self.spreadsheet_name = spreadsheet_name
        self.worksheet_name = worksheet_name
        self.sheet_owner = sheet_owner

        self.color_neutral = Color(1,1,1) # White
        self.color_warning = Color(1,0.5,0) # Orange
        self.color_alarm = Color(1,0,0) # Red

    def _collect_from_bucket(
            self,
            aws_access_key_id: str,
            aws_secret_access_key: str,
            aws_region: str,
            s3_path: str,
            aws_endpoint_url: str = None) -> BackupMetadata:

        kwargs = {
           "aws_access_key_id": aws_access_key_id,
           "aws_secret_access_key": aws_secret_access_key,
           "region_name": aws_region,
           "endpoint_url": aws_endpoint_url
        }

        s3 = boto3.resource(
            's3',
            **{k:v for k,v in kwargs.items() if v is not None}
        )

        logging.info(f"Collect metadata from {s3_path} ...")

        metadata_file_name = "/".join(s3_path.split("/")[3:])
        s3_path = s3_path.split("/")[2]
        metadata = s3.Object(s3_path, metadata_file_name).get()['Body'].read().decode("utf-8")
        metadata = json.loads(metadata)

        result = BackupMetadata()
        result.type = metadata.get("type", "None")
        result.size = metadata.get("size", "None")
        result.time = metadata.get("time", "None")
        result.customer = metadata.get("customer", "None")
        result.placement = metadata.get("placement", "None")
        result.backup_name = metadata.get("backup_name", "None")
        result.description = metadata.get("description", "None")
        result.count_of_backups = metadata.get("count_of_backups", "None")
        result.last_backup_date = metadata.get("last_backup_date", "None")
        result.supposed_backups_count = metadata.get("supposed_backups_count", "None")
        result.sha1sum = metadata.get("sha1sum", "None")
        for item in metadata.get("backups", []):
            result.backups.append(BackupFileInfo(
                backup_name=item.get("backup_name", "None"),
                size=item.get("size", "None"),
                backup_date=item.get("backup_date", "None"),
                sha1sum=item.get("sha1sum", "None")
            ))
        logging.info(f"Collect metadata from {s3_path} complete")
        return result

    def _csv_write(self, data: list, csv_path: str) -> None:
        with open(csv_path, 'a') as csvfile:
            csv_file = csv.writer(csvfile, delimiter=',', quotechar='|', quoting=csv.QUOTE_MINIMAL)
            for row in data:
                csv_file.writerow(row)

    def _compile_csv(self, metadata: list) -> str:
        logging.info(f"Compile csv file")
        csv_path = "tmp_report.csv"
        self._csv_write([[ "Customer", "DB type", "Backup Placement", "Size in MB", "Backup time spent", "Backup name", "Backup sha1 hash sum", "Backups count", "Supposed Backups Count", "Last Backup Date", "Description" ]], csv_path)

        backups_info = []
        for data in metadata:
            row = [ data.customer, data.type, data.placement, data.size, data.time, data.backup_name, data.sha1sum, data.count_of_backups, data.supposed_backups_count, data.last_backup_date, data.description ]
            backups_info.append(row)

        self._csv_write(backups_info, csv_path)

        return csv_path

    def _compile_file_backups_csv(self, metadata: list) -> str:
        logging.info(f"Compile csv file")
        csv_path = "tmp_report_file_backups.csv"
        self._csv_write([[ "Customer", "Description", "Backup name", "Backup sha1 hash sum", "Size in MB", "Backup Date" ]], csv_path)

        backups_info = []
        for data in metadata:
            if data.backups:
                for backup_file in data.backups:
                    row = [ data.customer, data.description, backup_file.backup_name, backup_file.sha1sum, backup_file.size, backup_file.backup_date ]
                    backups_info.append(row)

        self._csv_write(backups_info, csv_path)

        return csv_path

    def _upload_csv(self, csv_path: str, worksheet_name: str) -> None:
        logging.info(f"Upload csv to google sheet")
        scope = ["https://spreadsheets.google.com/feeds",
            "https://www.googleapis.com/auth/spreadsheets",
            "https://www.googleapis.com/auth/drive.file",
            "https://www.googleapis.com/auth/drive"]
        credentials = ServiceAccountCredentials.from_json_keyfile_name(self.credentials_path, scope)
        client = gspread.authorize(credentials)
        try:
            spreadsheet = client.open(self.spreadsheet_name)
            logging.debug("List current permissions")
            permissions = spreadsheet.list_permissions()
            logging.debug(permissions)
            for user in permissions:
                if user.get("emailAddress", None) == self.sheet_owner and user["role"] != "owner":
                    logging.info(f"Change owner to {self.sheet_owner}")
                    spreadsheet.transfer_ownership(user["id"])
                    break
        except gspread.exceptions.SpreadsheetNotFound as e:
            spreadsheet = client.create(self.spreadsheet_name)
            spreadsheet.share(self.sheet_owner, perm_type='user', role='writer')

        try:
            logging.debug(f"Worksheets are: {spreadsheet.worksheets()}")
            spreadsheet.worksheet(worksheet_name)
        except gspread.exceptions.WorksheetNotFound as e:
            spreadsheet.add_worksheet(title=worksheet_name, rows="100", cols="20")

        spreadsheet.values_clear(worksheet_name + "!A1:L10000")
        spreadsheet.values_update(
            worksheet_name,
            params={'valueInputOption': 'USER_ENTERED'},
            body={'values': list(csv.reader(open(csv_path)))}
        )

    def _get_backups_count(self, metadata: BackupMetadata) -> int:
        '''
            Return count of backups
        '''
        try:
            return int(metadata.count_of_backups)
        except ValueError as exc:
            # there are cases when total count string looks like "67 total / 10 full / 57 incremental"
            # so we need to parse it explicitly
            return int(metadata.count_of_backups.split(" ")[0])

    def _color_backup_count(self, metadata: BackupMetadata) -> Color:
        '''
            Select color for Backup count cell
        '''
        if self._get_backups_count(metadata) < 3:
            return self.color_alarm
        return self.color_neutral

    def _color_supposed_backups_count(self, metadata: BackupMetadata) -> Color:
        '''
            Select color for Supposed Backups Count
        '''
        if self._get_backups_count(metadata) <= int(metadata.supposed_backups_count) - 3:
            return self.color_alarm
        elif self._get_backups_count(metadata) <  int(metadata.supposed_backups_count) - 2:
            return self.color_warning
        return self.color_neutral

    def _color_last_backup_date(self, metadata: BackupMetadata) -> Color:
        '''
            Select color for Last Backup Date
        '''
        last_backup_date = dateparser.parse(metadata.last_backup_date)
        time_delta = datetime.datetime.now() - last_backup_date.replace(tzinfo=None)
        if time_delta.days > 7:
            return self.color_alarm
        return self.color_neutral

    def _set_color_matrix(self, metadata: list) -> list:
        '''
            Compile color matrix by collected metadata for google worksheet
        '''
        result = [[self.color_neutral, self.color_neutral, self.color_neutral, self.color_neutral, self.color_neutral]] # Worksheet header always white
        # Iterate over metadata, compile worksheet rows and colorize them
        for data in metadata:
            result.append([
                self.color_neutral, # Customer
                self.color_neutral, # DB type
                self.color_neutral, # Backup Placement
                self.color_neutral, # Size in MB
                self.color_neutral, # Backup time spent
                self.color_neutral, # Backup name
                self.color_neutral, # Backup hash sum
                self._color_backup_count(data), # Backup count
                self._color_supposed_backups_count(data), # Supposed Backups Count
                self._color_last_backup_date(data) , # Last Backup Date
                self.color_neutral, # Description
            ])

        return result

    def _get_column_name(self, n):
        '''
            Get letter from english alphabet by position number
        '''
        if n > 26:
            raise Exception("Function _get_column_name accepts a number from 0 to 26")

        result = ''
        while n > 0:
            index = (n - 1) % 26
            result += chr(index + ord('A'))
            n = (n - 1) // 26

        return result[::-1]

    def _colorize_worksheet(self, color_matrix: list) -> None:
        '''
            Colorize spreadsheet with colors sets in color_matrix
        '''
        scope = ["https://spreadsheets.google.com/feeds", 'https://www.googleapis.com/auth/spreadsheets', "https://www.googleapis.com/auth/drive.file", "https://www.googleapis.com/auth/drive"]
        credentials = ServiceAccountCredentials.from_json_keyfile_name(self.credentials_path, scope)
        spreadsheet = gspread.authorize(credentials).open(self.spreadsheet_name)
        worksheet = spreadsheet.worksheet(self.worksheet_name)

        # Drop all worksheet colors
        format_cell_range(
            worksheet=worksheet,
            name="0", # Set all cells for that operation
            cell_format=CellFormat(backgroundColor=Color(1, 1, 1)) # Colorize cells to white
        )

        # Iterate over color_matrix like over worksheet rows and its numbers
        for y, row in enumerate(color_matrix):
            # Iterate over worksheet cells in row and its column numbers
            for x, col in enumerate(row):
                if col == self.color_neutral: # Check if cell is white
                    continue
                sleep(5)
                format_cell_range(
                    worksheet=worksheet,
                    name=self._get_column_name(x+1)+str(y+1), # Compile cell name in format like "A1", "B2" etc
                    cell_format=CellFormat(backgroundColor=col)
                )

    def collect(self):
        metadata = []
        for bucket in self.buckets:
            metadata.append(
                self._collect_from_bucket(
                    aws_access_key_id=bucket.get("aws_access_key_id"),
                    aws_secret_access_key=bucket.get("aws_secret_access_key"),
                    aws_region=bucket.get("aws_region"),
                    s3_path=bucket.get("s3_path"),
                    aws_endpoint_url=bucket.get("aws_endpoint_url")
                )
            )

        csv = self._compile_csv(metadata)
        file_backups_csv = self._compile_file_backups_csv(metadata)
        self._upload_csv(csv, self.worksheet_name)
        self._upload_csv(file_backups_csv, self.worksheet_name + '_files')
        os.remove(csv)
        os.remove(file_backups_csv)

        color_matrix = self._set_color_matrix(metadata)
        self._colorize_worksheet(color_matrix)
